
#include "window.hpp"

#include <iostream>
#include <string>
#include <stdexcept>

#include <GLFW/glfw3.h>

using namespace fggl::gfx;

static void glfw_error(int code, const char* description) {
	std::cerr << "[GLFW] " << code << " " << description << std::endl;
}

static void framebuffer_resize(GLFWwindow* window, int width, int height) {
	glfwMakeContextCurrent( window );
	glViewport(0, 0, width, height);

	auto fgglWindow = reinterpret_cast<Window*>(glfwGetWindowUserPointer( window ));
	fgglWindow->framesize( width, height );
}

static void fggl_input_cursor(GLFWwindow* window, double x, double y) {
	Input& input = Input::instance();
	input.mousePos(x, y);
}

static void fggl_input_scroll(GLFWwindow* window, double x, double y) {
	Input& input = Input::instance();
	input.mouseScroll(x, y);
}

static void fggl_input_mouse_btn(GLFWwindow* window, int btn, int action, int mods) {
	// as we need the singleton for joysticks, might as well use it for these as well...
	Input& input = Input::instance();
	fggl::gfx::MouseButton buttonBit = (fggl::gfx::MouseButton)(1 << btn);
	input.mouseBtn( buttonBit, action == GLFW_PRESS );
}

static void fggl_joystick(int jid, int state) {
	Input& input = Input::instance();
	if ( state == GLFW_CONNECTED ) {
		fggl::gfx::Joystick data;
		data.name = glfwGetJoystickName(jid);
		data.axes = glfwGetJoystickAxes(jid, &data.axisCount);
		data.buttons = glfwGetJoystickButtons(jid, &data.buttonCount);
		data.hats = glfwGetJoystickHats(jid, &data.hatCount);

		input.joystickConnect(jid, data);
	} else if ( state == GLFW_DISCONNECTED ) {
		input.joystickDisconnect(jid);
	}
}

Context::Context() {
	glfwInitHint(GLFW_JOYSTICK_HAT_BUTTONS, GLFW_FALSE);

	int state = glfwInit();
	if ( state == GLFW_FALSE ) {
		const char** error;
		int code = glfwGetError(error);
		throw std::runtime_error( *error );
	}
}

Context::~Context() {
	glfwTerminate();
}

void Context::pollEvents() {
	glfwPollEvents();
}

Window::Window(Input& input) : m_window(nullptr), m_input(input) {
	glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 3);
	glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 3);
	glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE);
	glfwWindowHint(GLFW_OPENGL_DEBUG_CONTEXT, true);  

	m_window = glfwCreateWindow(1920, 1080, "main", nullptr, nullptr);
	if ( m_window == nullptr ) {
		return;
	}

	m_framesize = glm::vec2(1920, 1080);
	glfwSetWindowUserPointer(m_window, this);
	glfwSetFramebufferSizeCallback( m_window, framebuffer_resize );

	// input functions
	glfwSetScrollCallback(m_window, fggl_input_scroll);
	glfwSetCursorPosCallback(m_window, fggl_input_cursor);
	glfwSetMouseButtonCallback(m_window, fggl_input_mouse_btn);
	glfwSetJoystickCallback(fggl_joystick);
}

Window::~Window() {
	if ( m_window != nullptr ) {
		glfwDestroyWindow(m_window);
		m_window = nullptr;
	}
}

void Window::activate() const {
	assert( m_window != nullptr );
	glfwMakeContextCurrent(m_window);
}

void Window::swap() {
	assert( m_window != nullptr );
	glfwSwapBuffers(m_window);
}
